"""Concept factory."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Union, cast, get_args

from toisto.metadata import Language

from ..model_types import ConceptId
from .cefr import CommonReferenceLevel, CommonReferenceLevelSource
from .concept import Concept
from .grammar import GrammaticalCategory
from .label import label_factory

CommonReferenceLevelDict = dict[CommonReferenceLevel, CommonReferenceLevelSource | list[CommonReferenceLevelSource]]
ConceptIdListOrString = ConceptId | list[ConceptId]
ConceptIdDictOrListOrString = dict[Language, ConceptIdListOrString] | ConceptIdListOrString
MetaData = Literal["level", "antonym", "roots"]
LeafConceptDict = dict[
    Language | MetaData,
    ConceptId | list[ConceptId] | ConceptIdDictOrListOrString | CommonReferenceLevelDict,
]
CompositeConceptDict = dict[
    GrammaticalCategory | MetaData,
    Union["CompositeConceptDict", LeafConceptDict, CommonReferenceLevelDict],  # noqa: UP037
]
ConceptDict = LeafConceptDict | CompositeConceptDict


@dataclass
class ConceptFactory:
    """Create concepts from the concept dict."""

    concept_id: ConceptId
    concept_dict: ConceptDict

    def create_concept(self, parent_concept_id: ConceptId | None = None) -> Concept:
        """Create a concept from the concept_dict."""
        if self._get_grammatical_categories():
            return self._create_composite_concept(parent_concept_id)
        return self._create_leaf_concept(parent_concept_id)

    def _create_composite_concept(self, parent_concept_id: ConceptId | None = None) -> Concept:
        """Create a composite concept from a composite concept dict."""
        constituent_concept_ids = []
        roots_dict = cast(CompositeConceptDict, dict(roots=self._get_roots()))
        antonyms = self._get_antonym_concepts()
        levels = self._get_levels()
        for category in self._get_grammatical_categories():
            constituent_concept_id = ConceptId(f"{self.concept_id}/{category}")
            constituent_concept_ids.append(constituent_concept_id)
            antonym_concept_ids = [ConceptId(f"{antonym}/{category}") for antonym in antonyms]
            antonyms_dict = dict(antonym=antonym_concept_ids)
            constituent_concept_dict = (
                cast(CompositeConceptDict, self.concept_dict)[category] | roots_dict | antonyms_dict
            )
            constituent_concept_dict.setdefault("level", levels)
            concept_factory = self.__class__(constituent_concept_id, cast(ConceptDict, constituent_concept_dict))
            concept_factory.create_concept(self.concept_id)
        return Concept(
            self.concept_id,
            parent_concept_id,
            tuple(constituent_concept_ids),
            self._get_root_concepts(),
            self._get_antonym_concepts(),
            level=self._get_level(),
        )

    def _create_leaf_concept(self, parent_concept_id: ConceptId | None = None) -> Concept:
        """Create a leaf concept from a leaf concept dict."""
        labels = {
            cast(Language, key): label_factory(cast(str | list[str], value))
            for key, value in self.concept_dict.items()
            if key in get_args(Language)
        }
        return Concept(
            self.concept_id,
            parent_concept_id,
            (),
            self._get_root_concepts(),
            self._get_antonym_concepts(),
            labels,
            self._get_level(),
        )

    def _get_grammatical_categories(self) -> tuple[GrammaticalCategory, ...]:
        """Retrieve the grammatical categories from the concept dict."""
        return tuple(
            cast(GrammaticalCategory, key) for key in self.concept_dict if key in get_args(GrammaticalCategory)
        )

    def _get_root_concepts(self) -> dict[Language, tuple[ConceptId, ...]]:
        """Retrieve the roots from the concept dict."""
        roots = self._get_roots()
        if not isinstance(roots, dict):
            roots = {language: roots for language in get_args(Language)}  # Roots are the same for all languages
        return {
            language: tuple(concept_ids) if isinstance(concept_ids, list) else (concept_ids,)
            for language, concept_ids in roots.items()
        }

    def _get_roots(self) -> ConceptIdDictOrListOrString:
        """Get the roots from the concept dict."""
        return cast(ConceptIdDictOrListOrString, self.concept_dict.get("roots", {}))

    def _get_antonym_concepts(self) -> tuple[ConceptId, ...]:
        """Return the antonym concepts."""
        antonyms = self._get_antonyms()
        return tuple(antonyms) if isinstance(antonyms, list) else (antonyms,)

    def _get_antonyms(self) -> ConceptIdListOrString:
        """Get the antonyms from the concept dict."""
        return cast(ConceptIdListOrString, self.concept_dict.get("antonym", []))

    def _get_levels(self) -> CommonReferenceLevelDict:
        """Get the Common Reference Levels from the concept dict."""
        return cast(CommonReferenceLevelDict, self.concept_dict.get("level", {}))

    def _get_level(self) -> CommonReferenceLevel | None:
        """Determine the Common Reference Level for this concept.

        At the moment, just use the highest language level specified by the available sources.
        """
        concept_levels = [level for level in self._get_levels() if level in get_args(CommonReferenceLevel)]
        return max(concept_levels, default=None)



"""Labels."""

from __future__ import annotations


class Label(str):
    """Class representing labels for concepts."""

    def __eq__(self, other: object) -> bool:
        """Ignore hints when determining equality."""
        return self.split(";", maxsplit=1)[0] == str(other).split(";", maxsplit=1)[0]

    def __ne__(self, other: object) -> bool:
        """Return whether the labels are not equal."""
        return not self == other

    def __hash__(self) -> int:
        """Return the hash of the label, ignoring hints."""
        return hash(self.split(";", maxsplit=1)[0])

    @property
    def spelling_alternatives(self) -> Labels:
        """Extract the spelling alternatives from the label."""
        return tuple(self.__class__(label) for label in self.split(";", maxsplit=1)[0].split("|"))

    @property
    def hint(self) -> str:
        """Return the label hint, if any."""
        return self.split(";")[1] if ";" in self else ""


Labels = tuple[Label, ...]


def label_factory(string: str | list[str]) -> Labels:
    """Instantiate the labels from a string or list of strings."""
    labels = string if isinstance(string, list) else [string]
    return tuple(Label(label) for label in labels)



"""Quiz retention."""

from __future__ import annotations

from dataclasses import dataclass, fields
from datetime import datetime, timedelta

optional_datetime = datetime | None
SKIP_INTERVAL_GROWTH_FACTOR = 5  # Cf. https://artofmemory.com/blog/the-pimsleur-language-method/


@dataclass
class Retention:
    """Class to keep track of the retention of one quiz."""

    start: optional_datetime = None  # Start of the retention period, i.e. the period in which all answers were correct
    end: optional_datetime = None  # End of the retention period, i.e. the datetime of the most recent correct answer
    skip_until: optional_datetime = None  # Don't quiz this again until after the datetime
    count: int = 0  # Number of times the quiz was presented

    def increase(self) -> None:
        """Increase the retention of the quiz after a correct answer."""
        self.count += 1
        self.end = now = datetime.now()
        if self.start is None:
            self.start = now
        if self.count == 1:
            self.skip_until = now + timedelta(days=1)  # User already knows the answer, skip ahead
        elif now > self.start:
            self.skip_until = now + (now - self.start) * SKIP_INTERVAL_GROWTH_FACTOR
        else:
            self.skip_until = None

    def reset(self) -> None:
        """Reset the retention of the quiz after an incorrect answer."""
        self.count += 1
        self.skip_until = self.start = self.end = None

    @property
    def length(self) -> timedelta:
        """Return the length of the retention."""
        return self.end - self.start if self.start and self.end else timedelta()

    def is_silenced(self) -> bool:
        """Return whether the quiz is silenced."""
        return self.skip_until > datetime.now() if self.skip_until else False

    def as_dict(self) -> dict[str, str | int]:
        """Return the retention as dict."""
        result = {}
        for field in fields(self):
            key = field.name
            if value := getattr(self, key):
                result[key] = value.isoformat(timespec="seconds") if str(field.type) == "optional_datetime" else value
        return result

    @classmethod
    def from_dict(cls, retention_dict: dict[str, str | int]) -> Retention:
        """Instantiate a retention from a dict."""
        start = cls.__get_datetime(retention_dict, "start")
        end = cls.__get_datetime(retention_dict, "end")
        skip_until = cls.__get_datetime(retention_dict, "skip_until")
        count = int(retention_dict.get("count", 0))
        return cls(start, end, skip_until, count)

    @staticmethod
    def __get_datetime(retention_dict: dict[str, str | int], key: str) -> optional_datetime:
        """Get a datetime from the Retention dict."""
        return datetime.fromisoformat(str(retention_dict.get(key))) if retention_dict.get(key) else None



"""Meta data about this application."""

import pathlib
from importlib.metadata import metadata, version
from typing import Literal

import requests

_metadata = metadata("Toisto")
NAME = _metadata["name"]
SUMMARY = _metadata["summary"]
HOMEPAGE_URL = _metadata["Project-URL"].split(", ")[1]
CHANGELOG_URL = [url for url in _metadata.get_all("Project-URL") if "Changelog" in url][0].split(", ")[1]
TAGS_API_URL = "https://api.github.com/repos/fniessink/toisto/tags"

VERSION = version(NAME)
Language = Literal["en", "fi", "nl"]
SUPPORTED_LANGUAGES: dict[Language, str] = dict(en="English", fi="Finnish", nl="Dutch")

_topics_folder = pathlib.Path(__file__).parent.parent / "topics"
TOPIC_JSON_FILES = sorted(_topics_folder.glob("*.json"))
TOPICS = [json_file.stem for json_file in TOPIC_JSON_FILES]
PROGRESS_JSON = pathlib.Path.home() / f".{NAME.lower()}-progress.json"


def latest_version() -> str | None:
    """Return the latest version."""
    try:
        return requests.get(TAGS_API_URL, timeout=2).json()[0]["name"]
    except requests.ConnectionError:
        return None



"""Progress model class."""

from collections import deque

from toisto.model.language.concept import Concept

from .quiz import Quiz, Quizzes
from .retention import Retention
from .topic import Topics


class Progress:
    """Keep track of progress on quizzes."""

    def __init__(self, progress_dict: dict[str, dict[str, str | int]], topics: Topics) -> None:
        self.__progress_dict = {key: Retention.from_dict(value) for key, value in progress_dict.items()}
        self.__topics = topics
        self.__recent_concepts: deque[Concept] = deque(maxlen=2)  # Recent concepts to skip when selecting next quiz
        self.__quizzes_by_concept: dict[Concept, Quizzes] = {}
        for quiz in self.__topics.quizzes:
            self.__quizzes_by_concept.setdefault(quiz.concept.base_concept, set()).add(quiz)

    def increase_retention(self, quiz: Quiz) -> None:
        """Increase the retention of the quiz."""
        self.__progress_dict.setdefault(str(quiz), Retention()).increase()

    def reset_retention(self, quiz: Quiz) -> None:
        """Reset the retention of the quiz."""
        self.__progress_dict.setdefault(str(quiz), Retention()).reset()

    def next_quiz(self) -> Quiz | None:
        """Return the next quiz."""
        eligible_quizzes = {quiz for quiz in self.__topics.quizzes if self.__is_eligible(quiz)}
        quizzes_for_concepts_in_progress = {quiz for quiz in eligible_quizzes if self.__has_concept_in_progress(quiz)}
        quizzes_in_progress = {quiz for quiz in quizzes_for_concepts_in_progress if self.__in_progress(quiz)}
        for potential_quizzes in [quizzes_in_progress, quizzes_for_concepts_in_progress, eligible_quizzes]:
            unblocked_quizzes = self.__unblocked_quizzes(potential_quizzes, eligible_quizzes)
            if unblocked_quizzes:
                quiz = self.__sort_by_language_level(unblocked_quizzes)[0]
                self.__recent_concepts.append(quiz.concept.base_concept)
                return quiz
        return None

    def get_retention(self, quiz: Quiz) -> Retention:
        """Return the quiz retention."""
        return self.__progress_dict.get(str(quiz), Retention())

    def __is_eligible(self, quiz: Quiz) -> bool:
        """Return whether the quiz is not silenced and not the current quiz."""
        return quiz.concept.base_concept not in self.__recent_concepts and not self.get_retention(quiz).is_silenced()

    def __has_concept_in_progress(self, quiz: Quiz) -> bool:
        """Return whether the quiz's concept has been presented to the user before."""
        quizzes_for_same_concept = self.__quizzes_by_concept[quiz.concept.base_concept]
        return any(self.__in_progress(quiz_for_same_concept) for quiz_for_same_concept in quizzes_for_same_concept)

    def __in_progress(self, quiz: Quiz) -> bool:
        """Return whether the quiz has been presented to the user before."""
        return str(quiz) in self.__progress_dict

    def __unblocked_quizzes(self, quizzes: Quizzes, eligible_quizzes: Quizzes) -> Quizzes:
        """Return the quizzes that are not blocked by other quizzes.

        Quiz A is blocked by quiz B if the concept of quiz A is a compound with a root that is quizzed by quiz B.
        """
        return {
            quiz
            for quiz in quizzes
            if not self.__root_concepts_have_quizzes(quiz, eligible_quizzes)
            and not quiz.is_blocked_by(eligible_quizzes)
        }

    def __root_concepts_have_quizzes(self, quiz: Quiz, quizzes: Quizzes) -> bool:
        """Return whether the quiz's concept has root concepts that have quizzes."""
        return any(
            other_quiz
            for root in quiz.concept.root_concepts(quiz.question_language)
            for other_quiz in self.__quizzes_by_concept.get(root, set())
            if other_quiz != quiz and other_quiz in quizzes
        )

    def __sort_by_language_level(self, quizzes: Quizzes) -> list[Quiz]:
        """Sort the quizzes by the language level of the concept."""
        return sorted(quizzes, key=lambda quiz: str(quiz.concept.level))

    def as_dict(self) -> dict[str, dict[str, int | str]]:
        """Return the progress as dict."""
        return {key: value.as_dict() for key, value in self.__progress_dict.items()}



"""Command to show concepts."""

from itertools import chain

from rich.table import Table

from toisto.metadata import SUPPORTED_LANGUAGES, Language
from toisto.model.language.label import Labels
from toisto.model.quiz.topic import Topic, Topics
from toisto.ui.text import console


def enumerate_labels(labels: Labels) -> str:
    """Enumerate the labels."""
    return "\n".join(chain.from_iterable(label.spelling_alternatives for label in labels))


def topic_table(language: Language, source_language: Language, topic: Topic) -> Table:
    """Show the concepts of the topic."""
    table = Table(title=f"Topic {topic.name}")
    table.add_column(SUPPORTED_LANGUAGES[language])
    table.add_column(SUPPORTED_LANGUAGES[source_language])
    table.add_column("Grammatical categories")
    table.add_column("Language level")
    for concept in topic.concepts:
        for leaf_concept in concept.leaf_concepts():
            table.add_row(
                enumerate_labels(leaf_concept.labels(language)),
                enumerate_labels(leaf_concept.labels(source_language)),
                "/".join(leaf_concept.grammatical_categories()),
                leaf_concept.level,
            )
        table.add_section()
    return table


def show_topics(language: Language, source_language: Language, topics: Topics) -> None:
    """Show the concepts of the topics."""
    with console.pager():
        for topic in topics:
            console.print(topic_table(language, source_language, topic))



"""Grammar types."""

from typing import Literal

GrammaticalGender = Literal["female", "male", "neuter"]
GrammaticalNumber = Literal["infinitive", "singular", "plural"]
GrammaticalPerson = Literal["first person", "second person", "third person"]
DegreeOfComparison = Literal["positive degree", "comparative degree", "superlative degree"]
Tense = Literal["present tense", "past tense"]
SentenceForm = Literal["declarative", "interrogative"]
GrammaticalPolarity = Literal["affirmative", "negative"]
GrammaticalCategory = Literal[
    GrammaticalGender,
    GrammaticalNumber,
    GrammaticalPerson,
    DegreeOfComparison,
    Tense,
    SentenceForm,
    GrammaticalPolarity,
]



"""Quiz factory."""

from dataclasses import dataclass
from itertools import zip_longest

from toisto.metadata import Language

from ..language.concept import Concept
from .quiz import GRAMMATICAL_QUIZ_TYPES, Quiz, QuizType, Quizzes


@dataclass
class QuizFactory:
    """Create quizzes for concepts."""

    target_language: Language
    source_language: Language

    def create_quizzes(self, *concepts: Concept) -> Quizzes:
        """Create quizzes for the concepts."""
        return Quizzes().union(*(self.concept_quizzes(concept, Quizzes()) for concept in concepts))

    def concept_quizzes(self, concept: Concept, previous_quizzes: Quizzes) -> Quizzes:
        """Create the quizzes for a concept."""
        create_quizzes = self.composite_concept_quizzes if concept.constituent_concepts else self.leaf_concept_quizzes
        return create_quizzes(concept, previous_quizzes)

    def composite_concept_quizzes(self, concept: Concept, previous_quizzes: Quizzes) -> Quizzes:
        """Create the quizzes for a composite concept."""
        quizzes = Quizzes()
        for constituent_concept in concept.constituent_concepts:
            quizzes |= self.concept_quizzes(constituent_concept, quizzes.copy() | previous_quizzes)
        quizzes |= self.grammatical_quizzes(concept, quizzes.copy() | previous_quizzes)
        return quizzes

    def leaf_concept_quizzes(self, concept: Concept, previous_quizzes: Quizzes) -> Quizzes:
        """Create the quizzes for a leaf concept."""
        translation_quizzes = self.translation_quizzes(concept, previous_quizzes)
        listening_quizzes = self.listening_quizzes(concept, translation_quizzes.copy() | previous_quizzes)
        antonym_quizzes = self.antonym_quizzes(
            concept,
            translation_quizzes.copy() | listening_quizzes.copy() | previous_quizzes,
        )
        return translation_quizzes | listening_quizzes | antonym_quizzes

    def translation_quizzes(self, concept: Concept, previous_quizzes: Quizzes) -> Quizzes:
        """Create translation quizzes for the concept."""
        target_language, source_language = self.target_language, self.source_language
        target_labels, source_labels = concept.labels(target_language), concept.labels(source_language)
        if not target_labels or not source_labels:
            return Quizzes()
        blocked_by = tuple(previous_quizzes)
        target_to_source = Quizzes(
            Quiz(concept, target_language, source_language, target_label, source_labels, blocked_by=blocked_by)
            for target_label in target_labels
        )
        source_to_target = Quizzes(
            Quiz(concept, source_language, target_language, source_label, target_labels, blocked_by=blocked_by)
            for source_label in source_labels
        )
        return target_to_source | source_to_target

    def listening_quizzes(self, concept: Concept, previous_quizzes: Quizzes) -> Quizzes:
        """Create listening quizzes for the concept."""
        target_language, source_language = self.target_language, self.source_language
        labels = concept.labels(target_language)
        blocked_by = tuple(previous_quizzes)
        meanings = concept.meanings(source_language)
        return Quizzes(
            Quiz(concept, target_language, target_language, label, (label,), ("listen",), blocked_by, meanings)
            for label in labels
        )

    def grammatical_quizzes(self, concept: Concept, previous_quizzes: Quizzes) -> Quizzes:
        """Create grammatical quizzes for the concept."""
        target_language, source_language = self.target_language, self.source_language
        blocked_by = tuple(previous_quizzes)
        quizzes = Quizzes()
        for concept1, concept2 in concept.paired_leaf_concepts():
            labels1, labels2 = concept1.labels(target_language), concept2.labels(target_language)
            meanings = concept1.meanings(source_language) + concept2.meanings(source_language)
            quiz_types = grammatical_quiz_types(concept1, concept2)
            quizzes |= Quizzes(
                Quiz(concept, target_language, target_language, label1, (label2,), quiz_types, blocked_by, meanings)
                for label1, label2 in zip(labels1, labels2, strict=False)
                if label1 != label2
            )
        return quizzes

    def antonym_quizzes(self, concept: Concept, previous_quizzes: Quizzes) -> Quizzes:
        """Create antonym quizzes for the concept."""
        target_language, source_language = self.target_language, self.source_language
        labels = concept.labels(target_language)
        blocked_by = tuple(previous_quizzes)
        quizzes = Quizzes()
        for antonym in concept.antonym_concepts:
            antonym_labels = antonym.labels(target_language)
            meanings = concept.meanings(source_language) + antonym.meanings(source_language)
            quizzes |= Quizzes(
                Quiz(
                    concept,
                    target_language,
                    target_language,
                    label,
                    antonym_labels,
                    ("antonym",),
                    blocked_by,
                    meanings,
                )
                for label in labels
            )
        return quizzes


def grammatical_quiz_types(concept1: Concept, concept2: Concept) -> tuple[QuizType, ...]:
    """Return the quiz types to change the grammatical category of concept1 into that of concept2."""
    quiz_types = []
    for category1, category2 in zip_longest(concept1.grammatical_categories(), concept2.grammatical_categories()):
        if category1 != category2 and category2 is not None:
            quiz_types.append(GRAMMATICAL_QUIZ_TYPES[category2])
    return tuple(quiz_types)



"""Store and load progress data."""

from argparse import ArgumentParser
from typing import NoReturn

from ..metadata import NAME, PROGRESS_JSON
from ..model.quiz.progress import Progress
from ..model.quiz.topic import Topics
from .json_file import dump_json, load_json


def load_progress(topics: Topics, argument_parser: ArgumentParser) -> Progress | NoReturn:
    """Load the progress from the user's home folder."""
    try:
        progress_dict = load_json(PROGRESS_JSON, default={})
    except Exception as reason:  # noqa: BLE001
        return argument_parser.error(
            f"""{NAME} cannot parse the progress information in {PROGRESS_JSON}: {reason}.
To fix this, remove or rename {PROGRESS_JSON} and start {NAME} again. Unfortunately, this will reset your progress.
Please consider opening a bug report at https://github.com/fniessink/{NAME.lower()}. Be sure to attach the invalid
progress file to the issue.
""",
        )
    return Progress(progress_dict, topics)


def save_progress(progress: Progress) -> None:
    """Save the progress to the user's home folder."""
    dump_json(PROGRESS_JSON, progress.as_dict())



"""Text to speech."""

import os
import sys
import tempfile
from configparser import ConfigParser

import gtts
from playsound import playsound

SAY_VOICES = dict(en="Daniel", fi="Satu", nl="Xander")


def _say_with_macos_say(language: str, text: str, slow: bool) -> None:
    """Say the text with the MacOS say command."""
    voice = SAY_VOICES[language]
    rate = "--rate 150 " if slow else ""
    text = text.replace("'", r"\'")
    os.system(f"say --voice={voice} {rate}{text} &")


def _say_with_google_translate(language: str, text: str, config: ConfigParser, slow: bool) -> None:
    """Say the text with Google Translate say command."""
    try:
        service = gtts.gTTS(text=text, lang=language, slow=slow)
        with tempfile.NamedTemporaryFile(delete=False, suffix=".mp3") as mp3_file:
            service.save(mp3_file.name)
    except gtts.tts.gTTSError as reason:
        message = f"Can't use Google Translate: {reason}"
        raise RuntimeError(message) from reason
    mp3_play_command = config.get("commands", "mp3player")
    if mp3_play_command == "playsound":
        playsound(mp3_file.name)
    else:
        os.system(f"{mp3_play_command} {mp3_file.name} &")


def say(language: str, text: str, config: ConfigParser, slow: bool = False) -> None:
    """Say the text in the specified language."""
    try:
        _say_with_google_translate(language, text, config, slow)
    except RuntimeError:
        if sys.platform == "darwin":
            _say_with_macos_say(language, text, slow)
        else:
            raise



"""Common European Framework of Reference for Languages (CEFR)."""

from typing import Literal

# Source: https://www.coe.int/en/web/common-european-framework-reference-languages/level-descriptions
CommonReferenceLevel = Literal["A1", "A2", "B1", "B2", "C1", "C2"]

# Toisto uses different sources to assess the common reference level for concepts:
CommonReferenceLevelSource = Literal["EP", "KK", "OD"]

SOURCES: dict[CommonReferenceLevelSource, dict[Literal["name", "url", "language"], str]] = dict(
    EP=dict(
        name="English Vocabulary Profile Online - British English",
        url="https://www.englishprofile.org/wordlists/evp",
        language="en",
    ),
    KK=dict(name="Yle Kielikoulu Learning Profile", url="https://kielikoulu.yle.fi/#/profile", language="fi"),
    OD=dict(
        name="Oxford Advanced Learner's Dictionary online",
        url="https://www.oxfordlearnersdictionaries.com",
        language="en",
    ),
)



"""Concept classes."""

from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass, field
from itertools import permutations
from typing import ClassVar, cast, get_args

from toisto.metadata import Language
from toisto.tools import zip_and_cycle

from ..model_types import ConceptId
from .cefr import CommonReferenceLevel
from .grammar import GrammaticalCategory
from .label import Label, Labels


@dataclass
class Concept:
    """Class representing language concepts.

    A concept is either a composite or a leaf concept. Composite concepts have have two or more constituent concepts,
    representing different grammatical categories, for example singular and plural forms. Leaf concepts have labels
    in different languages.
    """

    concept_id: ConceptId
    _parent_concept: ConceptId | None = None
    _constituent_concepts: tuple[ConceptId, ...] = ()
    _root_concepts: dict[Language, tuple[ConceptId, ...]] = field(default_factory=dict)
    _antonym_concepts: tuple[ConceptId, ...] = ()
    _labels: dict[Language, Labels] = field(default_factory=dict)
    level: CommonReferenceLevel | None = None

    instances: ClassVar[dict[ConceptId, Concept]] = {}

    def __post_init__(self) -> None:
        """Add the concept to the concept id -> concept mapping."""
        self.instances[self.concept_id] = self

    def __hash__(self) -> int:
        """Return the concept hash."""
        return hash(self.concept_id)

    def leaf_concepts(self) -> Iterable[Concept]:
        """Return this concept's leaf concepts, or self if this concept is a leaf concept."""
        if self.constituent_concepts:
            for concept in self.constituent_concepts:
                yield from concept.leaf_concepts()
        else:
            yield self

    def paired_leaf_concepts(self) -> Iterable[tuple[Concept, Concept]]:
        """Pair the leaf concepts from the constituent concepts."""
        for concept_group in zip_and_cycle(*[list(concept.leaf_concepts()) for concept in self.constituent_concepts]):
            for permutation in permutations(concept_group, r=2):
                yield cast(tuple[Concept, Concept], permutation)

    @property
    def constituent_concepts(self) -> tuple[Concept, ...]:
        """Return the constituent concepts of this concept."""
        return self._get_concepts(self._constituent_concepts)

    @property
    def base_concept(self) -> Concept:
        """Return the base concept of this concept."""
        return self.instances[self._parent_concept].base_concept if self._parent_concept in self.instances else self

    def root_concepts(self, language: Language) -> tuple[Concept, ...]:
        """Return the root concepts of this compound concept, for the specified language."""
        return self._get_concepts(self._root_concepts.get(language, ()))

    @property
    def antonym_concepts(self) -> tuple[Concept, ...]:
        """Return the antonym concepts of this concept."""
        return self._get_concepts(self._antonym_concepts)

    def labels(self, language: Language) -> Labels:
        """Return the labels for the language."""
        return self._labels.get(language, Labels())

    def meanings(self, language: Language) -> Labels:
        """Return the meaning of the concept in the specified language."""
        meaning = Label(self._labels.get(language, [""])[0])
        return (meaning,) if meaning else ()

    def grammatical_categories(self) -> tuple[GrammaticalCategory, ...]:
        """Return the grammatical categories of this concept."""
        keys = self.concept_id.split("/")
        return tuple(cast(GrammaticalCategory, key) for key in keys if key in get_args(GrammaticalCategory))

    def _get_concepts(self, concept_ids: tuple[ConceptId, ...]) -> tuple[Concept, ...]:
        """Return the concepts with the given concept ids."""
        return tuple(self.instances[concept_id] for concept_id in concept_ids if concept_id in self.instances)



"""Command to show progress information."""

from dataclasses import dataclass
from datetime import datetime
from typing import Literal

from rich.table import Table

from toisto.metadata import SUPPORTED_LANGUAGES, Language
from toisto.model.quiz.progress import Progress
from toisto.model.quiz.quiz import Quiz
from toisto.model.quiz.topic import Topics
from toisto.ui.format import format_datetime, format_duration
from toisto.ui.text import console

SortColumn = Literal["attempts", "retention"]
RETENTION_ATTRIBUTE = dict(attempts="count", retention="length")


@dataclass
class QuizSorter:
    """Class to provide a sort function to sort quizzes."""

    progress: Progress
    sort: SortColumn

    def get_sort_key(self, quiz: Quiz) -> str:
        """Return the retention attribute to sort by."""
        return getattr(self.progress.get_retention(quiz), RETENTION_ATTRIBUTE[self.sort])


def show_progress(language: Language, topics: Topics, progress: Progress, sort: SortColumn = "attempts") -> None:
    """Show progress."""
    table = Table(title=f"Progress {SUPPORTED_LANGUAGES[language]}")
    table.add_column("Quiz type")
    table.add_column("Question")
    table.add_column("From")
    table.add_column("To")
    table.add_column("Answer(s)")
    table.add_column("Attempts", justify="right")
    table.add_column("Retention")
    table.add_column("Not quizzed until")
    sorted_quizzes = sorted(topics.quizzes, key=QuizSorter(progress, sort).get_sort_key, reverse=True)
    for quiz in sorted_quizzes:
        retention = progress.get_retention(quiz)
        skip = retention.skip_until
        quiz_types = " and ".join(quiz.quiz_types)
        table.add_row(
            quiz_types.capitalize(),
            quiz.question,
            quiz.question_language,
            quiz.answer_language,
            "\n".join(quiz.answers),
            str(retention.count),
            format_duration(retention.length) if retention.length else "",
            format_datetime(skip) if skip and skip > datetime.now() else "",
        )
    with console.pager():
        console.print(table)



"""Formatting."""

from datetime import datetime, timedelta


def format_duration(duration: timedelta) -> str:
    """Format the duration in a human friendly way."""
    if duration.days > 1:
        return f"{duration.days} days"
    seconds = duration.seconds + (24 * 3600 if duration.days else 0)
    if seconds >= 1.5 * 3600:  # 1.5 hours
        hours = round(seconds / 3600)
        return f"{hours} hours"
    if seconds >= 1.5 * 60:  # 1.5 minutes
        minutes = round(seconds / 60)
        return f"{minutes} minutes"
    return f"{seconds} seconds"


def format_datetime(date_time: datetime) -> str:
    """Return a human readable version of the datetime."""
    return date_time.isoformat(sep=" ", timespec="minutes")



"""Config file parser."""

import logging
import sys
from configparser import ConfigParser, Error
from pathlib import Path


def read_config() -> ConfigParser:
    """Read the config file."""
    parser = ConfigParser()
    try:
        with Path("~/.toisto.cfg").expanduser().open("r", encoding="utf-8") as config_file:
            parser.read_file(config_file)
    except (OSError, Error) as reason:
        logging.warning("Could not read ~/.toisto.cfg: %s", reason)
    if parser.get("commands", "mp3player", fallback=None) is None:
        if "commands" not in parser.sections():
            parser.add_section("commands")
        defaults = dict(darwin="afplay", linux="mpg123 --quiet")
        parser["commands"]["mp3player"] = defaults.get(sys.platform, "playsound")
    return parser



"""Zip function that cycles shorter lists."""

from collections.abc import Iterator
from itertools import cycle
from typing import TypeVar

T = TypeVar("T")


def zip_and_cycle(*lists: list[T]) -> Iterator[tuple[T]]:
    """Zip the lists, while cycling lists that are shorter than the longest list."""
    max_len = max(len(lst) for lst in lists) if lists else 0
    cycled_lists = [cycle(lst) if len(lst) < max_len else lst for lst in lists]
    yield from zip(*cycled_lists, strict=False)



"""Link words to an online dictionary."""

from string import punctuation

DICTIONARY_URL = "https://en.wiktionary.org/wiki"


def linkify_and_enumerate(*texts: str, sep: str = ", ") -> str:
    """Return a linkified and enumerated version of the texts."""
    return sep.join(f'"{linkify(text)}"' for text in texts)


def linkify(text: str) -> str:
    """Return a version of the text where each word is turned into a link to a dictionary."""
    linkified_words = []
    for word in text.split():
        prefix, word, postfix = split_punctuation(word)
        linkified_words.append(f"{prefix}[link={DICTIONARY_URL}/{word.lower()}]{word}[/link]{postfix}")
    return " ".join(linkified_words)


def split_punctuation(text: str) -> tuple[str, str, str]:
    """Return a tuple of prefixed punctuation, the text without punctuation, and postfix punctuation."""
    stripped_text = text.strip(punctuation)
    stripped_text_start = text.find(stripped_text)
    stripped_text_end = stripped_text_start + len(stripped_text)
    return text[:stripped_text_start], text[stripped_text_start:stripped_text_end], text[stripped_text_end:]



"""Main module for the application."""

import logging
from contextlib import suppress

with suppress(ImportError):
    import readline  # noqa: F401 `readline` imported but unused

# Suppress warning messages printed by the playsound module.
logging.getLogger().setLevel(logging.ERROR)

from .command.practice import practice
from .command.show_progress import show_progress
from .command.show_topics import show_topics
from .metadata import latest_version
from .persistence.config import read_config
from .persistence.progress import load_progress
from .persistence.topics import load_topics
from .ui.cli import create_argument_parser
from .ui.text import show_welcome


def main() -> None:
    """Run the main program."""
    config = read_config()
    argument_parser = create_argument_parser(config)
    args = argument_parser.parse_args()
    topics = load_topics(
        args.target_language,
        args.source_language,
        args.levels,
        args.topic,
        args.topic_file,
        argument_parser,
    )
    progress = load_progress(topics, argument_parser)
    if args.command == "practice":
        show_welcome(latest_version())
        practice(progress, config)
    elif args.command == "topics":
        show_topics(args.target_language, args.source_language, topics)
    else:
        show_progress(args.target_language, topics, progress, args.sort)



"""Load concepts from topic files and generate quizzes."""

import pathlib
from argparse import ArgumentParser
from typing import NoReturn

from ..metadata import NAME, TOPIC_JSON_FILES, Language
from ..model.language.cefr import CommonReferenceLevel
from ..model.language.concept_factory import ConceptFactory
from ..model.quiz.quiz_factory import QuizFactory
from ..model.quiz.topic import Topic, Topics
from .json_file import load_json


def load_topics(  # noqa: PLR0913
    target_language: Language,
    source_language: Language,
    levels: list[CommonReferenceLevel],
    builtin_topics_to_load: list[str],
    topic_files_to_load: list[str],
    argument_parser: ArgumentParser,
) -> Topics | NoReturn:
    """Load the topics with the concepts and generate the quizzes."""
    topics = set()
    topic_files: list[pathlib.Path] = []
    if builtin_topics_to_load or topic_files_to_load:
        topic_files.extend(topic for topic in TOPIC_JSON_FILES if topic.stem in builtin_topics_to_load)
        topic_files.extend(pathlib.Path(topic) for topic in topic_files_to_load)
    else:
        topic_files.extend(TOPIC_JSON_FILES)
    quiz_factory = QuizFactory(target_language, source_language)
    for topic_file in topic_files:
        concepts = []
        topic_quizzes = set()
        try:
            for concept_key, concept_value in load_json(topic_file).items():
                concept = ConceptFactory(concept_key, concept_value).create_concept()
                if levels and concept.level not in levels:
                    continue
                concepts.append(concept)
                topic_quizzes.update(quiz_factory.create_quizzes(concept))
        except Exception as reason:  # noqa: BLE001
            return argument_parser.error(f"{NAME} cannot read topic {topic_file}: {reason}.\n")
        topics.add(Topic(topic_file.stem, tuple(concepts), topic_quizzes))
    return Topics(topics)



"""Quiz classes."""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from itertools import chain
from typing import Literal, cast

from toisto.metadata import SUPPORTED_LANGUAGES, Language

from ..language.concept import Concept
from ..language.grammar import GrammaticalCategory
from ..language.label import Label, Labels
from .match import match

QuizType = Literal[
    "translate",
    "listen",
    "antonym",
    "pluralize",
    "singularize",
    "masculinize",
    "feminize",
    "neuterize",
    "give positive degree",
    "give comparative degree",
    "give superlative degree",
    "give first person",
    "give second person",
    "give third person",
    "give infinitive",
    "give present tense",
    "give past tense",
    "make declarative",
    "make interrogative",
    "affirm",
    "negate",
]
GRAMMATICAL_QUIZ_TYPES: dict[GrammaticalCategory, QuizType] = {
    "plural": "pluralize",
    "singular": "singularize",
    "male": "masculinize",
    "female": "feminize",
    "neuter": "neuterize",
    "positive degree": "give positive degree",
    "comparative degree": "give comparative degree",
    "superlative degree": "give superlative degree",
    "first person": "give first person",
    "second person": "give second person",
    "third person": "give third person",
    "infinitive": "give infinitive",
    "present tense": "give present tense",
    "past tense": "give past tense",
    "declarative": "make declarative",
    "interrogative": "make interrogative",
    "affirmative": "affirm",
    "negative": "negate",
}
QUIZ_TYPE_GRAMMATICAL_CATEGORIES = {value: key for key, value in GRAMMATICAL_QUIZ_TYPES.items()}
INSTRUCTIONS: dict[QuizType, str] = dict(
    translate="Translate into",
    listen="Listen and write in",
    antonym="Give the [underline]antonym[/underline] in",
)


def instruction(*quiz_types: QuizType) -> str:
    """Return the instruction text for the quiz types."""
    if instruction_label := INSTRUCTIONS.get(quiz_types[0]):
        return instruction_label
    categories = " ".join(QUIZ_TYPE_GRAMMATICAL_CATEGORIES[quiz_type] for quiz_type in quiz_types)
    return f"Give the [underline]{categories}[/underline] in"


@dataclass(frozen=True)
class Quiz:
    """Class representing a quiz."""

    concept: Concept
    question_language: Language
    answer_language: Language
    _question: Label
    _answers: Labels
    quiz_types: tuple[QuizType, ...] = ("translate",)
    blocked_by: tuple[Quiz, ...] = ()
    _meanings: Labels = Labels()
    _hash: int = 0

    def __post_init__(self) -> None:
        """Initialize calculated attributes."""
        # The dataclass is frozen, so some magic is needed to set the hash attribute.
        super().__setattr__(
            "_hash",
            hash((self.question_language, self.answer_language, self.question, self.quiz_types)),
        )

    def __str__(self) -> str:
        """Return a string version of the quiz that can be used as key in the progress dict."""
        quiz_types = "+".join(self.quiz_types)
        return f"{self.question_language}:{self.answer_language}:{self._question}:{quiz_types}"

    def __hash__(self) -> int:
        """Return a hash using the same attributes as used for testing equality."""
        return self._hash

    def __eq__(self, other: object) -> bool:
        """Return whether this quiz is equal to the other."""
        if not isinstance(other, self.__class__):
            return False
        return (
            self.question_language == other.question_language
            and self.answer_language == other.answer_language
            and self.question == other.question
            and self.quiz_types == other.quiz_types
        )

    def __ne__(self, other: object) -> bool:
        """Return whether this quiz is not equal to the other."""
        return not self == other

    def is_correct(self, guess: Label) -> bool:
        """Return whether the guess is correct."""
        return match(guess, *self.answers)

    @cached_property
    def question(self) -> Label:
        """Return the first spelling alternative of the question."""
        return self._question.spelling_alternatives[0]

    @property
    def answer(self) -> Label:
        """Return the first spelling alternative of the first answer."""
        return self._answers[0].spelling_alternatives[0]

    @property
    def answers(self) -> Labels:
        """Return all answers."""
        answers = [answer.spelling_alternatives for answer in self._answers]
        return cast(Labels, tuple(chain(*answers)))

    @property
    def meanings(self) -> Labels:
        """Return the first spelling alternative of the meanings."""
        return Labels(meaning.spelling_alternatives[0] for meaning in self._meanings)

    def other_answers(self, guess: Label) -> Labels:
        """Return the answers not equal to the guess."""
        if self.quiz_types == ("listen",):
            return Labels()  # Other answers doesn't make sense if the user has to type what is spoken
        return tuple(answer for answer in self.answers if not match(guess, answer))

    def instruction(self) -> str:
        """Generate the quiz instruction."""
        hint = self._question.hint
        hint = f" ({hint})" if self.question_language != self.answer_language and hint else ""
        return f"{instruction(*self.quiz_types)} {SUPPORTED_LANGUAGES[self.answer_language]}{hint}"

    def is_blocked_by(self, quizzes: Quizzes) -> bool:
        """Return whether this quiz should come after any of the given quizzes."""
        return bool(set(self.blocked_by) & quizzes)


Quizzes = set[Quiz]



"""Command-line interface."""

from argparse import ArgumentParser, _SubParsersAction
from configparser import ConfigParser
from typing import get_args

from rich_argparse import RichHelpFormatter

from ..metadata import HOMEPAGE_URL, SUMMARY, SUPPORTED_LANGUAGES, TOPICS, VERSION, latest_version
from ..model.language.cefr import CommonReferenceLevel


def add_language_arguments(parser: ArgumentParser, config: ConfigParser) -> None:
    """Add the language arguments to the parser."""
    choices = SUPPORTED_LANGUAGES.keys()
    for argument in ("target", "source"):
        default = config.get("languages", argument, fallback=None)
        default_help = f"default: {default}; " if default else ""
        parser.add_argument(
            f"-{argument[0]}",
            f"--{argument}",
            choices=choices,
            default=default,
            dest=f"{argument}_language",
            help=f"{argument} language; {default_help}available languages: %(choices)s",
            metavar="{language}",
            required=not default,
        )


def add_topic_arguments(parser: ArgumentParser) -> None:
    """Add the topic arguments to the parser."""
    parser.add_argument(
        "-T",
        "--topic",
        action="append",
        default=[],
        choices=TOPICS,
        metavar="{topic}",
        help="topic to use, can be repeated; default: all; available topics: %(choices)s",
    )
    parser.add_argument(
        "-f",
        "--topic-file",
        action="append",
        default=[],
        metavar="{topic file}",
        help="topic file to use, can be repeated",
    )


def add_level_arguments(parser: ArgumentParser, config: ConfigParser) -> None:
    """Add the language level arguments to the parser."""
    levels = get_args(CommonReferenceLevel)
    default = [level for level in levels if level in config.get("languages", "levels", fallback="")]
    default_help = ", ".join(default) if default else "all"
    parser.add_argument(
        "-l",
        "--level",
        action="append",
        default=default,
        dest="levels",
        metavar="{level}",
        choices=levels,
        help=f"language levels to use, can be repeated; default: {default_help}; available levels: %(choices)s",
    )


def add_practice_command(subparser: _SubParsersAction, config: ConfigParser) -> None:
    """Add a practice command."""
    command_help = (
        "practice a language, for example type `%(prog)s practice --target fi --source en` to "
        "practice Finnish from English"
    )
    add_command(subparser, "practice", "Practice a language.", command_help, config)


def add_progress_command(subparser: _SubParsersAction, config: ConfigParser) -> None:
    """Add a command to show progress."""
    command_help = (
        "show progress, for example `%(prog)s progress --target fi --source en` shows progress "
        "on practicing Finnish from English"
    )
    parser = add_command(subparser, "progress", "Show progress.", command_help, config)
    parser.add_argument(
        "-S",
        "--sort",
        metavar="{option}",
        choices=["retention", "attempts"],
        default="retention",
        help="how to sort progress information; default: by retention; available options: %(choices)s",
    )


def add_topics_command(subparser: _SubParsersAction, config: ConfigParser) -> None:
    """Add a command to show topics."""
    command_help = "show topics, for example `%(prog)s topics --topic nature` shows the contents of the nature topic"
    add_command(subparser, "topics", "Show topics.", command_help, config)


def add_command(
    subparser: _SubParsersAction,
    command: str,
    description: str,
    command_help: str,
    config: ConfigParser,
) -> ArgumentParser:
    """Add a command."""
    parser = subparser.add_parser(
        command,
        description=description,
        help=command_help,
        formatter_class=RichHelpFormatter,
    )
    add_language_arguments(parser, config)
    add_level_arguments(parser, config)
    add_topic_arguments(parser)
    return parser


def create_argument_parser(config: ConfigParser) -> ArgumentParser:
    """Create the argument parser."""
    epilog = f"See {HOMEPAGE_URL}/blob/main/README.md for more information."
    argument_parser = ArgumentParser(description=SUMMARY, epilog=epilog, formatter_class=RichHelpFormatter)
    latest = latest_version()
    version = f"v{VERSION}" + (f" ({latest} is available)" if latest and latest.strip("v") > VERSION else "")
    argument_parser.add_argument("-V", "--version", action="version", version=version)
    command_help = "type `%(prog)s {command} --help` for more information on a command"
    subparser_action = argument_parser.add_subparsers(
        dest="command",
        title="commands",
        help=command_help,
        required=True,
    )
    add_practice_command(subparser_action, config)
    add_progress_command(subparser_action, config)
    add_topics_command(subparser_action, config)
    return argument_parser



"""Practice command."""

from configparser import ConfigParser

from toisto.model.language.label import Label
from toisto.model.quiz.progress import Progress
from toisto.model.quiz.quiz import Quiz
from toisto.persistence.progress import save_progress
from toisto.ui.dictionary import linkify
from toisto.ui.speech import say
from toisto.ui.text import DONE, TRY_AGAIN, console, feedback_correct, feedback_incorrect, instruction


def do_quiz_attempt(quiz: Quiz, config: ConfigParser, attempt: int = 1) -> tuple[Label, bool]:
    """Present the question, get the answer from the user, and evaluate it."""
    while True:
        say(quiz.question_language, quiz.question, config, slow=attempt > 1)
        if answer := Label(input("> ").strip()):
            break
        print("\033[F", end="")  # noqa: T201  # Move cursor one line up
    return answer, quiz.is_correct(answer)


def do_quiz(quiz: Quiz, progress: Progress, config: ConfigParser) -> None:
    """Do one quiz and update the progress."""
    console.print(instruction(quiz))
    if "listen" not in quiz.quiz_types:
        console.print(linkify(quiz.question))
    answer, correct = do_quiz_attempt(quiz, config)
    if not correct and answer != "?":
        console.print(TRY_AGAIN)
        answer, correct = do_quiz_attempt(quiz, config, attempt=2)
    if correct:
        progress.increase_retention(quiz)
        feedback = feedback_correct(answer, quiz)
    else:
        progress.reset_retention(quiz)
        feedback = feedback_incorrect(answer, quiz)
    console.print(feedback)


def practice(progress: Progress, config: ConfigParser) -> None:
    """Practice a language."""
    try:
        while quiz := progress.next_quiz():
            do_quiz(quiz, progress, config)
        console.print(DONE)
    except (KeyboardInterrupt, EOFError):
        console.print()  # Make sure the shell prompt is displayed on a new line
    finally:
        save_progress(progress)



"""Topic class."""

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass, field
from itertools import chain

from ..language.concept import Concept
from .quiz import Quizzes


@dataclass(frozen=True)
class Topic:
    """Collection of quizzes for concepts centered around a topic."""

    name: str
    concepts: tuple[Concept, ...]
    quizzes: Quizzes

    def __hash__(self) -> int:
        """Return the hash of the file name."""
        return hash(self.name)


@dataclass
class Topics:
    """Collection of topics."""

    topics: set[Topic] = field(default_factory=set)

    @property
    def quizzes(self) -> Quizzes:
        """Return all quizzes."""
        return set(chain.from_iterable(topic.quizzes for topic in self.topics))

    def __iter__(self) -> Iterator[Topic]:
        """Return an iterator over self.topics."""
        return iter(self.topics)
